from pathlib import Path
from schema import Schema, And, Use, SchemaError
from src import pistachio


import json
import pytest
import shutil


# Schema ----------------------------------------------------------------------
TREE_SCHEMA = Schema({
    "path": And(Use(str)),
    "results": And(Use(list))
})


DESCRIBE_SCHEMA = Schema({
    "path": And(Use(str)),
    "exists": And(Use(bool)),
    "is_directory": And(Use(bool)),
    "is_file": And(Use(bool)),
    "is_symlink": And(Use(bool)),
    "name": And(Use(str)),
    "stem": And(Use(str)),
    "suffix": And(Use(str))
})


def schema_validation(example, schema):
    """
    Method to confirm with a dictionary matches a schema.
    """
    try:
        schema.validate(example)
        return True
    except SchemaError:
        return False


# State -----------------------------------------------------------------------
def setup_module():
    """
    Setup the required resources neccessary to run all the tests.
    """
    with open("example.txt", "w") as fh:
        fh.write("The quick brown fox jumps over the lazy dog\n")
        fh.close()

    Path("./tests/abc").mkdir()
    Path("./tests/abc/xyz").mkdir()

    Path("./tests/abc/file-1.txt").touch()
    Path("./tests/abc/file-2.txt").symlink_to("./tests/abc/file-1.txt")
    Path("./tests/abc/xyz/file-3.txt").touch()
    Path("./tests/abc/xyz/file-4.txt").touch()

    Path("./456").mkdir()
    Path("./456/789").mkdir()
    Path("./456/file-8.txt").touch()
    Path("./456/file-9.txt").symlink_to("./tests/abc/file-1.txt")
    Path("./file-6.txt").touch()
    Path("./file-7.txt").symlink_to("./tests/abc/file-1.txt")


def teardown_module():
    """
    Remove anything generated by testing.
    """
    Path("README.doc").unlink()
    Path("example.rtf").unlink()
    Path("example.txt").unlink()

    shutil.rmtree("./tests/abc/")
    shutil.rmtree("./tests/123/")


# Fixtures --------------------------------------------------------------------
@pytest.fixture
def tree_expected_results():
    return json.dumps(
        {
            "path": "./tests/abc",
            "results": [
                {
                    "path": "./def",
                    "exists": True,
                    "is_directory": True,
                    "is_file": False,
                    "is_symlink": False,
                    "name": "def",
                    "stem": "def",
                    "suffix": None
                },
                {
                    "path": "./file-1.txt",
                    "exists": True,
                    "is_directory": False,
                    "is_file": True,
                    "is_symlink": False,
                    "name": "file-1.txt",
                    "stem": "file-1",
                    "suffix": "txt"
                },
                {
                    "path": "./file-2.txt",
                    "exists": True,
                    "is_directory": False,
                    "is_file": False,
                    "is_symlink": True,
                    "name": "file-2.txt",
                    "stem": "file-2",
                    "suffix": "txt"
                },
                {
                    "path": "./file-6.txt",
                    "exists": True,
                    "is_directory": False,
                    "is_file": True,
                    "is_symlink": False,
                    "name": "file-6.txt",
                    "stem": "file-6",
                    "suffix": "txt"
                },
                {
                    "path": "./file-7.txt",
                    "exists": True,
                    "is_directory": False,
                    "is_file": False,
                    "is_symlink": True,
                    "name": "file-7.txt",
                    "stem": "file-7",
                    "suffix": "txt"
                },
                {
                    "path": "./ghi",
                    "exists": True,
                    "is_directory": True,
                    "is_file": False,
                    "is_symlink": False,
                    "name": "ghi",
                    "stem": "ghi",
                    "suffix": None
                },
                {
                    "path": "./ghi/jkl",
                    "exists": True,
                    "is_directory": True,
                    "is_file": False,
                    "is_symlink": False,
                    "name": "jkl",
                    "stem": "jkl",
                    "suffix": None
                },
                {
                    "path": "./xyz",
                    "exists": True,
                    "is_directory": True,
                    "is_file": False,
                    "is_symlink": False,
                    "name": "xyz",
                    "stem": "xyz",
                    "suffix": None
                },
                {
                    "path": "./xyz/file-3.txt",
                    "exists": True,
                    "is_directory": False,
                    "is_file": True,
                    "is_symlink": False,
                    "name": "file-3.txt",
                    "stem": "file-3",
                    "suffix": "txt"
                },
                {
                    "path": "./xyz/file-4.txt",
                    "exists": True,
                    "is_directory": False,
                    "is_file": True,
                    "is_symlink": False,
                    "name": "file-4.txt",
                    "stem": "file-4",
                    "suffix": "txt"
                }
            ]
        },
        indent=2,
        sort_keys=True
    )


# Tests -----------------------------------------------------------------------
def test_copy_file():
    """
    Test the method to copy and paste a file on the filesystem.
    """
    result = pistachio.cp("example.txt", "example.rtf")

    assert result is True
    assert Path("example.rtf").exists() is True


def test_copy_folder():
    """
    Test the method to copy and paste a file on the filesystem.
    """
    result = pistachio.cp("./tests/abc", "./tests/123")

    assert result is True
    assert Path("./tests/123").exists() is True


def test_copy_symlink():
    """
    Test the method to copy and paste a file on the filesystem.
    """
    result = pistachio.cp("./tests/abc/file-2.txt", "./tests/123/file-5.txt")

    assert result is True
    assert Path("./tests/123/file-5.txt").is_symlink() is True


def test_describe_schema_directory():
    """
    Test to validate the describe method response for a directory.
    """
    example = pistachio.describe("tests")
    assert schema_validation(example, DESCRIBE_SCHEMA) is True


def test_describe_schema_file():
    """
    Test to validate the describe method response for a file.
    """
    example = pistachio.describe("README.md")
    assert schema_validation(example, DESCRIBE_SCHEMA) is True


def test_describe_schema_symlink():
    """
    Test to validate the describe method response for a symbolic link.
    """
    example = pistachio.describe("README.rst")
    assert schema_validation(example, DESCRIBE_SCHEMA) is True


def test_exists_true():
    """
    Test to confirm the exists method returns True.
    """
    assert pistachio.exists("README.md") is True


def test_exists_false():
    """
    Test to confirm the exists method returns False.
    """
    assert pistachio.exists("LICENSE.md") is False


def test_get_md5_hash():
    """
    Test to confirm the get_md5_hash method returns False.
    """
    md5_hash_str = pistachio.get_md5_hash("example.txt")
    assert md5_hash_str == "37c4b87edffc5d198ff5a185cee7ee09"


def test_get_md5_hash_returns_none():
    """
    Test to confirm the get_md5_hash method returns None.
    """
    assert pistachio.get_md5_hash("example.docx") is None


def test_is_directory_true():
    """
    Test to confirm the is_directory method returns True.
    """
    assert pistachio.is_directory("tests") is True


def test_is_directory_false():
    """
    Test to confirm the is_directory method returns False.
    """
    assert pistachio.is_directory("LICENSE.md") is False


def test_is_file_true():
    """
    Test to confirm the is_file method returns True.
    """
    assert pistachio.is_file("README.md") is True


def test_is_file_false():
    """
    Test to confirm the is_file method returns False.
    """
    assert pistachio.is_file("tests") is False


def test_make_directory():
    """
    Method to verify that a directory can be created.
    """
    path_str = "./tests/abc/def"
    pistachio.mkdir(path_str)

    assert Path(path_str).exists() is True


def test_make_directory_recursively():
    """
    Method to verify that a directory path can be created recursively.
    """
    path_str = "./tests/abc/ghi/jkl"
    pistachio.mkdir(path_str)

    assert Path(path_str).exists() is True


def test_mv_directory():
    """
    Method to confirm that a directory can be moved.
    """
    result = pistachio.mv("./456", "./tests/123/456")

    assert result is True


def test_mv_file():
    """
    Method to confirm that a file can be moved.
    """
    result = pistachio.mv("./file-6.txt", "./tests/abc/file-6.txt")

    assert result is True


def test_mv_symlink():
    """
    Method to confirm that a symlink can be moved.
    """
    result = pistachio.mv("./file-7.txt", "./tests/abc/file-7.txt")

    assert result is True


def test_name_example_directory():
    """
    Test to confirm the name method returns "None".
    """
    assert pistachio.name("./foo/example") == "example"


def test_name_example_file():
    """
    Test to confirm the name method returns "example.txt".
    """
    assert pistachio.name("./foo/example.txt") == "example.txt"


def test_stem_example_directory():
    """
    Test to confirm the stem method returns "example".
    """
    assert pistachio.stem("./foo/example") == "example"


def test_stem_example_file():
    """
    Test to confirm the stem method returns "example".
    """
    assert pistachio.stem("./foo/example.txt") == "example"


def test_suffix_example_directory():
    """
    Test to confirm the suffix method returns "None".
    """
    assert pistachio.suffix("./foo/example") is None


def test_suffix_example_file():
    """
    Test to confirm the suffix method returns ".txt".
    """
    assert pistachio.suffix("./foo/example.txt") == "txt"


def test_touch_new_file_true():
    """
    Test to confirm the touch method returns True.
    """
    assert pistachio.touch("README.doc") is True


def test_touch_file_exist_returns_false():
    """
    Test to confirm the touch method returns True.
    """
    assert pistachio.touch("README.md") is False


def test_touch_directory_does_not_exist_returns_error():
    """
    Test to confirm the touch method raised FileNotFoundError exception.
    """
    with pytest.raises(FileNotFoundError):
        pistachio.touch("docs/README.doc")


def test_tree():
    """
    Test to confirm that the tree method returns a list of dictionaries.
    """
    r = pistachio.tree("tests/abc")

    assert schema_validation(r, TREE_SCHEMA) is True
    assert all(
        schema_validation(i, DESCRIBE_SCHEMA) is True for i in r["results"]
    )


def test_tree_results(tree_expected_results):
    """
    Test to confirm that the tree method returns a list of dictionaries.
    """
    results = json.dumps(
        pistachio.tree("./tests/abc"),
        indent=2,
        sort_keys=True
    )

    assert tree_expected_results == results


def test_tree_no_directory():
    """
    Test to confirm the tree method raised FileNotFoundError exception.
    """
    with pytest.raises(FileNotFoundError):
        pistachio.tree("docs")



from pathlib import Path


import errno
import hashlib
import os
import shutil


def cp(source_path_str, target_path_str):
    """
    Method to copy and paste a resource from one location to another.
    """
    if is_directory(source_path_str):
        shutil.copytree(
            source_path_str,
            target_path_str,
            symlinks=True,
            copy_function=shutil.copy
        )

    if is_file(source_path_str):
        shutil.copy(
            source_path_str,
            target_path_str
        )

    if is_symlink(source_path_str):
        Path(target_path_str).symlink_to(os.readlink(source_path_str))

    return exists(target_path_str)


def describe(path_str):
    """
    Method to describe the type of resources.
    """
    return {
        "path": path_str,
        "exists": exists(path_str),
        "is_directory": is_directory(path_str),
        "is_file": is_file(path_str),
        "is_symlink": is_symlink(path_str),
        "name": name(path_str),
        "stem": stem(path_str),
        "suffix": suffix(path_str)
    }


def exists(path_str):
    """
    Method to return True or False whether a resource exists.
    """
    if is_symlink(path_str):
        return True
    else:
        return Path(path_str).exists()


def get_md5_hash(path_str):
    """
    Method to return the MD5 hash of a file.
    """
    if exists(path_str) is True and is_file(path_str) is True:
        md5_hash = hashlib.md5()
        with open(path_str, "rb") as fh:
            for block in iter(lambda: fh.read(4096), b""):
                md5_hash.update(block)
            fh.close()
        return md5_hash.hexdigest()
    else:
        return None


def is_directory(path_str):
    """
    Method to return True or False whether a resource is a directory.
    """
    return Path(path_str).is_dir()


def is_file(path_str):
    """
    Method to return True or False whether a resource is a file.
    """
    return Path(path_str).is_file()


def is_symlink(path_str):
    """
    Method to return True or False whether a resource is a symbolic link.
    """
    return Path(path_str).is_symlink()


def mkdir(path_str):
    """
    Method to create a new directory or directories recursively.
    """
    return Path(path_str).mkdir(parents=True, exist_ok=True)


def mv(source_path_str, target_path_str):
    """
    Method to move a resource from one location to another.
    """
    shutil.move(source_path_str, target_path_str)

    return exists(target_path_str)


def name(path_str):
    """
    Method to return the name of a resource.
    """
    return os.path.basename(path_str)


def stem(path_str):
    """
    Return the stem of the last item in the path.
    """
    return Path(path_str).stem


def suffix(path_str):
    """
    Return the file extension suffix of the last item in the path.
    """
    suffix = Path(path_str).suffix

    clean = [
        (".", "")
    ]

    for old, new in clean:
        suffix = suffix.replace(old, new)

    return suffix if suffix != '' else None


def touch(path_str):
    """
    Method to generated an empty file.
    """
    if exists(path_str) is False:
        try:
            open(path_str, "a").close()
            return True
        except FileNotFoundError:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), path_str
            )
    else:
        return False


def tree(path_str):
    """
    Method to walk through a directory tree and discover all files
    and directories on the file system.
    """
    results_lst = []

    if exists(path_str) and is_directory(path_str):
        initial_path_str = os.getcwd()
        os.chdir(path_str)

        for root, directories, filenames in os.walk("."):
            for directory in directories:
                results_lst.append(
                    describe(f"""{root}/{directory}""")
                )
            for filename in filenames:
                results_lst.append(
                    describe(f"""{root}/{filename}""")
                )

        os.chdir(initial_path_str)

        return {
            "path": path_str,
            "results": sorted(results_lst, key=lambda d: d['path'])
        }
    else:
        raise FileNotFoundError(
            errno.ENOENT, os.strerror(errno.ENOENT), path_str
        )



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pistachio",
    version="0.4.2",
    description="Pistachio aims to simplify reoccurring tasks when working with the file system.",
    py_modules=["pistachio"],
    package_dir={"": "src"},
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: OS Independent",
    ],
    url="https://github.com/mjakinowittering/pistachio",
    author="Matthew Akino-Wittering",
    author_email="matthew.akinowittering@gmail.com",
    install_requires = [],
    extras_require = {
        "dev": [
            "pytest >= 3.7",
            "flake8",
            "check-manifest",
            "schema",
            "twine",
        ],
    },
)
